package me.icymint.libra.sage.core.dialect;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import me.icymint.libra.sage.core.dialect.query.ConstQueryVar;
import me.icymint.libra.sage.core.dialect.query.QueryCondition;
import me.icymint.libra.sage.core.dialect.query.QueryVar;
import me.icymint.libra.sage.core.query.Execute;
import me.icymint.libra.sage.core.query.Query;
import me.icymint.libra.sage.core.result.MetaDataResultHandler;
import me.icymint.libra.sage.core.result.ScalarListResult;
import me.icymint.libra.sage.core.result.ScalarResult;
import me.icymint.libra.sage.model.SqlFunction;
import me.icymint.libra.sage.model.SqlProcedure;
import me.icymint.libra.sage.model.SqlTable;
import me.icymint.libra.sage.model.SqlType;
import me.icymint.libra.sage.model.SqlVar;
import me.icymint.libra.sage.model.SqlVarType;
import me.icymint.libra.sage.model.format.BeautyFormat;
import me.icymint.libra.sage.model.format.FormatSql;

public abstract class AbstractDialect implements SqlDialect {

	/**
	 * 函数结果查询。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-9下午2:09:42
	 * 
	 */
	private class QueryFunctions extends ScalarListResult<String> implements
			MetaDataResultHandler<List<String>> {
		private final String defaultCatalog;

		QueryFunctions(String defaultCatalog) {
			super(3, 1);
			this.defaultCatalog = defaultCatalog;
		}

		@Override
		public String handle(ResultSet rs, int collumnIndex)
				throws SQLException {
			return getFunctionName(super.handle(rs, collumnIndex),
					rs.getString(2), rs.getInt(8));
		}

		@Override
		public ResultSet queryResultSet(DatabaseMetaData metadata)
				throws SQLException {
			return metadata.getProcedures(defaultCatalog, null, null);
		}
	}

	/**
	 * 存储过程结果查询。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-9下午2:10:10
	 * 
	 */
	private class QueryProcedures extends ScalarListResult<String> implements
			MetaDataResultHandler<List<String>> {

		private String defaultCatalog;

		QueryProcedures(String defaultCatalog) {
			super(3, 1);
			this.defaultCatalog = defaultCatalog;
		}

		@Override
		public String handle(ResultSet rs, int collumnIndex)
				throws SQLException {
			return getProcedureName(super.handle(rs, collumnIndex),
					rs.getString(2), rs.getInt(8));
		}

		@Override
		public ResultSet queryResultSet(DatabaseMetaData metadata)
				throws SQLException {
			return metadata.getProcedures(defaultCatalog, null, null);
		}
	}

	/**
	 * 查询表格信息。
	 * 
	 * @author Daniel
	 * @datetime 2012-11-9上午11:54:10
	 * 
	 * @param <T>
	 */
	private class QueryTables extends ScalarListResult<String> implements
			MetaDataResultHandler<List<String>> {

		private String[] types;
		private final String defaultCatalog;

		/**
		 * 构建查询表格的执行指令。
		 * 
		 * @param rh
		 *            结果处理类。
		 * @param types
		 *            表格类型，请参考
		 *            {@link DatabaseMetaData#getTables(String ,String ,String , String[])}
		 */
		QueryTables(String[] types, String defaultCatalog) {
			super(3, 1);
			this.defaultCatalog = defaultCatalog;
			this.types = types;
		}

		@Override
		public String handle(ResultSet rs, int collumnIndex)
				throws SQLException {
			return super.handle(rs, collumnIndex).toLowerCase();
		}

		@Override
		public ResultSet queryResultSet(DatabaseMetaData metadata)
				throws SQLException {
			return metadata.getTables(defaultCatalog, null, null, types);
		}

	}

	protected final static String ls = ",\n	";

	private final String name;

	private final static Collection<Integer> auto_increment = new LinkedHashSet<Integer>();

	static {
		auto_increment.add(Types.BIGINT);
		auto_increment.add(Types.INTEGER);
		auto_increment.add(Types.NUMERIC);
	}
	private final Map<SqlVarType, String> jdbc = new LinkedHashMap<SqlVarType, String>();

	private final Map<String, SqlVarType> fjdbc = new LinkedHashMap<String, SqlVarType>();
	{
		for (SqlVarType svt : SqlVarType.values()) {
			fjdbc.put(svt.name(), svt);
		}
		initialTypeMap(jdbc);
		for (SqlVarType svt : jdbc.keySet()) {
			fjdbc.put(jdbc.get(svt), svt);
		}
	}

	/**
	 * JDBC连接URL的正则表达式匹配。
	 */
	private static final Pattern url = Pattern
			.compile("jdbc\\:([_a-z][_0-9a-z]*)\\:\\/\\/[^/]+?\\/([_a-z][_0-9a-z]*)");

	protected AbstractDialect(String name) {
		this.name = name;
	}

	protected void execute(String sql, Connection conn) throws SQLException {
		new Execute(sql).query(conn);
	}

	@Override
	public String executeFunc(SqlFunction sf) {
		if (sf.getVarSize() == 0)
			return "SELECT " + sqlPrefix() + sf.toString() + "();";
		String abc = "";
		for (int i = 1; i < sf.getVarSize(); i++)
			abc = abc + ",?";
		return "SELECT " + sqlPrefix() + sf.toString() + "(?" + abc.trim()
				+ ");";
	}

	@Override
	public String executeProc(SqlProcedure sp) {
		if (sp.getVarSize() == 0)
			return "{CALL " + sqlPrefix() + sp.toString() + "()}";
		String abc = "";
		for (int i = 1; i < sp.getVarSize(); i++)
			abc = abc + ",?";
		return "{CALL " + sqlPrefix() + sp.toString() + "(?" + abc.trim()
				+ ")}";
	}

	@Override
	public String getAdvancedQuerySql(SqlTable t, QueryVar[] select,
			QueryCondition where, Map<SqlTable, String> map, SqlVar[] group,
			SqlVar[] order, int limit, int start) throws SQLSyntaxException {
		int count = 0;
		String sql = "";
		if (select != null) {
			for (QueryVar v : select) {
				if (v.getTable() != null && v.getTable() != t
						&& !map.containsKey(v.getTable()))
					throw new SQLSyntaxException();
				if (!(v instanceof ConstQueryVar))
					count++;
			}
			sql += "SELECT " + select[0].getText();
			for (int i = 1; i < select.length; i++) {
				sql += "," + select[i].getText();
			}
		} else {
			count++;
			sql += "SELECT *";
		}
		if (count > 0) {
			sql += " FROM " + getTableName(t);
			for (SqlTable tt : map.keySet()) {
				sql += " " + map.get(tt) + " " + getTableName(tt);
			}
			if (where != null)
				sql += " WHERE " + where.getText();
			sql += list(" GROUP BY", ",", group);
			sql += list(" ORDER BY", ",", group);
			if (limit > 0)
				sql += " LIMIT " + limit;
		}
		return sql;
	}

	@Override
	public String getCreateDatabase(String database) {
		return "CREATE DATABASE " + database;
	}

	@Override
	public String getCurrentCatalog(Connection conn) throws SQLException {
		return conn.getCatalog();
	}

	@Override
	public String getDatabaseProductName() {
		return name;
	}

	@Override
	public String getDefaultCatalog(Connection conn) throws SQLException {
		Matcher m = url.matcher(conn.getMetaData().getURL());
		if (m.find()) {
			return m.group(2);
		} else
			throw new SQLException("默认数据库未指定，且无法自动获取默认数据库名称");
	}

	@Override
	public String getDeleteByNames(SqlTable t, SqlVar[] names) {
		String sql = "DELETE FROM " + this.sqlPrefix() + t.toString();
		return sql + where(names);
	}

	@Override
	public String getDeleteDatabase(String database) {
		return "DROP DATABASE " + database;
	}

	@Override
	public String getDeleteFunctionSql(SqlFunction sf) {
		return "DROP FUNCTION " + sf.toString();
	}

	@Override
	public String getDeleteProcedureSql(SqlProcedure sp) {
		return "DROP PROCEDURE " + sp.toString();
	}

	@Override
	public String getDeleteTableSql(SqlTable st) {
		return "DROP TABLE " + st.toString();
	}

	@Override
	public FormatSql getFormat() {
		return new BeautyFormat();
	}

	/**
	 * 获取函数名称。 由于不同的JdbcDriver会返回不同的函数信息。 因此需要该方法来设立方言处理。
	 * 
	 * @param function
	 * @param schema
	 * @param type
	 */
	protected abstract String getFunctionName(String function, String schema,
			int type);

	@Override
	public String getInsert(SqlTable t) {
		String sql = "INSERT INTO " + this.sqlPrefix() + t.toString()
				+ " VALUES(?";
		int num = t.getVarSize();
		if (t.isAutoIncrement()) {
			num = num - t.getPrimaryKeys().length;
		}
		for (int i = 1; i < num; i++) {
			sql += ",?";
		}
		return sql + ")";
	}

	/**
	 * 获取存储过程名称。 由于不同的JdbcDriver会返回不同的存储过程信息。 因此需要该方法来设立方言处理。
	 * 
	 * @param procedure
	 * @param schema
	 * @param type
	 */
	protected abstract String getProcedureName(String procedure, String schema,
			int type);

	@Override
	public String getQueryAll(SqlTable t) {
		return "SELECT * FROM " + this.sqlPrefix() + t.toString();
	}

	@Override
	public String getQueryByNames(SqlTable t, SqlVar[] names) {
		String sql = "SELECT * FROM " + this.sqlPrefix() + t.toString();
		return sql + where(names);
	}

	@Override
	public MetaDataResultHandler<List<String>> getQueryFunctions(
			String defaultCatalog) {
		return new QueryFunctions(defaultCatalog);
	}

	@Override
	public MetaDataResultHandler<List<String>> getQueryProcedures(
			String defaultCatalog) {
		return new QueryProcedures(defaultCatalog);
	}

	@Override
	public MetaDataResultHandler<List<String>> getQueryTables(String[] types,
			String defaultCatalog) {
		return new QueryTables(types, defaultCatalog);
	}

	@Override
	public String getTableName(SqlTable t) {
		return sqlPrefix() + t.toString();
	}

	@Override
	public String getUpdate(SqlTable t, SqlVar[] v, SqlVar[] key) {
		String sql = "UPDATE " + this.sqlPrefix() + t.toString();
		return sql + set(v) + where(key);
	}

	@Override
	public String getVarName(SqlVar v) {
		return varPrefix() + v.toString();
	}

	public String getVarName(SqlVar v, SqlTable t) {
		return t.isMask() ? v.toString() : v.getName();
	}

	protected abstract void initialTypeMap(Map<SqlVarType, String> jdbc);

	@Override
	public boolean isAutoIncrement(SqlType st) {
		return auto_increment.contains(st.getJdbcType());
	}

	private String list(String head, String midfix, SqlVar[] names) {
		String sql = "";
		if (names.length == 0) {
			return sql;
		} else
			sql += " " + head + " " + names[0].toString() + "=? ";
		for (int i = 1; i < names.length; i++) {
			sql += midfix + " " + names[i].toString() + "=? ";
		}
		return sql;
	}

	protected String query(String sql, Connection conn) throws SQLException {
		return new Query<String>(sql, new ScalarResult<String>(), 0)
				.query(conn);
	}

	private String set(SqlVar[] names) {
		return list("SET", ",", names);
	}

	@Override
	public void setCurrentCatalog(Connection conn, String cat)
			throws SQLException {
		conn.setCatalog(cat);
	}

	@Override
	public String sqlPrefix() {
		return "";
	}

	@Override
	public SqlVarType transfer(int sqlType) {
		for (SqlVarType svt : SqlVarType.values()) {
			if (svt.getIndex() == sqlType)
				return svt;
		}
		throw new NullPointerException("序号为" + sqlType + "的JDBC类型未找到！");
	}

	@Override
	public String transfer(SqlVarType jdbcType) {
		if (jdbc.containsKey(jdbcType))
			return jdbc.get(jdbcType);
		else
			return jdbcType.name();
	}

	@Override
	public String varPrefix() {
		return "";
	}

	private String where(SqlVar[] names) {
		return list("WHERE", "AND", names);
	}
}
