package me.icymint.libra.sage.core.dialect;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;

import me.icymint.libra.sage.core.dialect.query.QueryCondition;
import me.icymint.libra.sage.core.dialect.query.QueryVar;
import me.icymint.libra.sage.core.result.MetaDataResultHandler;
import me.icymint.libra.sage.model.SqlFunction;
import me.icymint.libra.sage.model.SqlProcedure;
import me.icymint.libra.sage.model.SqlTable;
import me.icymint.libra.sage.model.SqlType;
import me.icymint.libra.sage.model.SqlVar;
import me.icymint.libra.sage.model.SqlVarType;
import me.icymint.libra.sage.model.format.FormatSql;
import me.icymint.libra.sage.model.format.SqlFormatException;

public interface SqlDialect {
	/**
	 * 
	 * @param sf
	 * @return 执行自定义函数指令。
	 */
	String executeFunc(SqlFunction sf);

	/**
	 * 执行存储过程指令。
	 * 
	 * @param sp
	 * @return
	 */
	String executeProc(SqlProcedure sp);

	/**
	 * 高级查询。
	 * 
	 * @param t
	 * @param select
	 * @param where
	 * @param map
	 * @param group
	 * @param order
	 * @param limit
	 * @param start
	 * @throws SQLSyntaxException
	 */
	String getAdvancedQuerySql(SqlTable t, QueryVar[] select,
			QueryCondition where, Map<SqlTable, String> map, SqlVar[] group,
			SqlVar[] order, int limit, int start) throws SQLSyntaxException;

	/**
	 * 创建数据库。
	 * 
	 * @param database
	 */
	String getCreateDatabase(String database);

	/**
	 * 创建函数语句。
	 * 
	 * @param sf
	 * @throws SqlFormatException
	 */
	String getCreateFunctionSql(SqlFunction sf) throws SqlFormatException;

	/**
	 * 创建存储过程语句。
	 * 
	 * @param sp
	 * @throws SqlFormatException
	 */
	String getCreateProcedureSql(SqlProcedure sp) throws SqlFormatException;

	/**
	 * 创建表格语句。
	 * 
	 * @param t
	 * @throws SqlFormatException
	 */
	String getCreateTableSql(SqlTable t) throws SqlFormatException;

	String getCurrentCatalog(Connection conn) throws SQLException;

	/**
	 * 数据库产品名称。
	 */
	String getDatabaseProductName();

	String getDefaultCatalog(Connection conn) throws SQLException;

	String getDeleteByNames(SqlTable t, SqlVar[] names);

	/**
	 * 删除数据库命令。
	 * 
	 * @param database
	 * @return
	 */
	String getDeleteDatabase(String database);

	/**
	 * 删除函数语句。
	 * 
	 * @param sf
	 */
	String getDeleteFunctionSql(SqlFunction sf);

	/**
	 * 删除函数语句。
	 * 
	 * @param sp
	 */
	String getDeleteProcedureSql(SqlProcedure sp);

	/**
	 * 删除表格语句。
	 * 
	 * @param st
	 */
	String getDeleteTableSql(SqlTable st);

	// /**
	// * 获取函数名称。 由于不同的JdbcDriver会返回不同的函数信息。 因此需要该方法来设立方言处理。
	// *
	// * @param function
	// * @param schema
	// * @param type
	// */
	// String getFunctionName(String function, String schema, int type);

	/**
	 * 
	 * @return 获取方言名称。
	 */
	String getDialectName();

	// /**
	// * 获取存储过程名称。 由于不同的JdbcDriver会返回不同的存储过程信息。 因此需要该方法来设立方言处理。
	// *
	// * @param procedure
	// * @param schema
	// * @param type
	// */
	// String getProcedureName(String procedure, String schema, int type);

	/**
	 * 获取输出SQL的格式化类。
	 */
	FormatSql getFormat();

	/**
	 * 插入数据。
	 * 
	 * @param t
	 */
	String getInsert(SqlTable t);

	/**
	 * 获取表格的所有数据。
	 * 
	 * @param t
	 */
	String getQueryAll(SqlTable t);

	/**
	 * 获取表格某些特征的数据。
	 * 
	 * @param t
	 * @param names
	 */
	String getQueryByNames(SqlTable t, SqlVar[] names);

	MetaDataResultHandler<List<String>> getQueryFunctions(String defaultCatalog);

	MetaDataResultHandler<List<String>> getQueryProcedures(String defaultCatalog);

	/**
	 * 获取查询表格的接口。
	 * 
	 * @param types
	 * @param defaultCatalog
	 */
	MetaDataResultHandler<List<String>> getQueryTables(String[] types,
			String defaultCatalog);

	/**
	 * 获取特定数据库的特定数据表名称。
	 * 
	 * @param t
	 */
	String getTableName(SqlTable t);

	String getUpdate(SqlTable t, SqlVar[] v, SqlVar[] key);

	/**
	 * 获取特定数据库的特定字段名称。
	 * 
	 * @param v
	 */
	String getVarName(SqlVar v);

	/**
	 * 确定哪些字段可以被设置为自增。
	 * 
	 * @param sqlType
	 */
	boolean isAutoIncrement(SqlType sqlType);

	boolean isSupportAutoCreateDatabase();

	void setCurrentCatalog(Connection conn, String cat) throws SQLException;

	/**
	 * 
	 * @return 获取sql方言的特殊前缀。
	 */
	String sqlPrefix();

	SqlVarType transfer(int sqlType);

	/**
	 * 把JDBC Type转换成数据库对于的字段。
	 * 
	 * @param jdbcType
	 */
	String transfer(SqlVarType jdbcType);

	/**
	 * 
	 * @return 存储过程和函数内部参数调用的特殊前缀。
	 */
	String varPrefix();
}
