package me.icymint.libra.sage.model;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.sql.SQLException;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicInteger;

import javax.sql.DataSource;

import me.icymint.libra.sage.core.JdbcUtils;
import me.icymint.libra.sage.core.dialect.Dialect;
import me.icymint.libra.sage.model.ddl.FetchTable;
import me.icymint.libra.sage.model.format.SqlFormatException;
import me.icymint.libra.sage.model.operator.SQLDuplicateException;
import me.icymint.libra.sage.model.operator.SQLExistsException;
import me.icymint.libra.sage.model.operator.SQLNotExistsException;
import me.icymint.libra.sage.model.orm.Mapping;
import me.icymint.libra.sage.model.orm.MappingException;
import me.icymint.libra.sage.model.orm.MappingImpl;
import me.icymint.libra.sage.model.xml.SqlParser;
import me.icymint.libra.xml.XmlParserException;

/**
 * SQL工厂方法类，其集成了所有SQL对象的创建方法，以及提供了一个SQL仓库用于存放这些创建以后的对象。
 * 其行为更接近与Builder，并且一个工厂类的实例只能存放一个SQL仓库，该仓库自动创建。
 * 
 * @author Daniel
 * @datetime 2012-12-19下午2:07:25
 * 
 */
public class SqlFactory {
	private static AtomicInteger sfid = new AtomicInteger();
	private static Map<Integer, SqlFactory> map = new LinkedHashMap<Integer, SqlFactory>();
	private static Map<String, SqlFactory> nmap = new LinkedHashMap<String, SqlFactory>();

	public static SqlFactory cloneInstance(DataSource ds, String databasename,
			SqlFactory sf) throws FileNotFoundException, XmlParserException,
			SQLException {
		return cloneInstance(ds, databasename, null, sf);
	}

	public static SqlFactory cloneInstance(DataSource ds, String databasename,
			String prefix, SqlFactory sf) throws FileNotFoundException,
			XmlParserException, SQLException {
		return cloneInstance(getSqlInfo(ds, databasename, prefix), sf);
	}

	public static SqlFactory cloneInstance(SqlInfo si, SqlFactory sf)
			throws FileNotFoundException, XmlParserException, SQLException {
		File f;
		File temp = new File(System.getProperty("java.io.tmpdir"));
		// File temp = new File("F:");
		do {
			f = new File(temp, UUID.randomUUID().toString());
		} while (f.canWrite());
		sf.save(new FileOutputStream(f));
		f.setReadOnly();
		SqlFactory sx = SqlFactory.newInstance(si, new FileInputStream(f));
		f.delete();
		sx.getDatabase().rebuild();
		return sx;
	}

	public static Set<String> fetchAllRegistedNames() {
		return nmap.keySet();
	}

	/**
	 * 根据数据库名称，获取SQL工厂对象。
	 * 
	 * @param name
	 * @throws SQLException
	 *             当SQL工厂不存在或者重复的时候，抛出。
	 */
	public static SqlFactory fetchByDbname(String name) throws SQLException {
		SqlFactory sf = null;
		for (SqlFactory f : map.values()) {
			if (f.getDatabase().getName().equals(name)) {
				if (sf != null)
					throw new SQLDuplicateException("SQL工厂实例存在两个数据库名称为" + name
							+ "的SQL工厂");
				else
					sf = f;
			}
		}
		if (sf == null)
			throw new SQLNotExistsException("无法找到数据库名称为" + name + "的SQL工厂实例");
		else
			return sf;
	}

	/**
	 * 通过ID进行搜索。
	 * 
	 * @param id
	 */
	public static SqlFactory fetchById(int id) {
		return map.get(id);
	}

	/**
	 * 通过注册名称，查找SQL工厂实例。 需要自行注册。
	 * 
	 * @param name
	 * @see #registerFactory(SqlFactory, String)
	 */
	public static SqlFactory fetchByRegistedName(String name) {
		return nmap.get(name);
	}

	/**
	 * 从数据库中自动抓取并生成SQL工厂类实例。
	 * 
	 * @param ds
	 *            数据源
	 * @param databasename
	 *            数据库名称
	 * @throws SQLException
	 */
	public static SqlFactory fetchInstance(DataSource ds, String databasename)
			throws SQLException {
		return fetchInstance(ds, databasename, null);
	}

	/**
	 * 从数据库中自动抓取并生成SQL工厂类实例。
	 * 
	 * @param ds
	 *            数据源
	 * @param databasename
	 *            数据库名称
	 * @param prefix
	 *            数据库内SQL对象使用的前缀，该前缀用于表名，存储过程名和函数名称。
	 * @throws SQLException
	 */
	public static SqlFactory fetchInstance(DataSource ds, String databasename,
			String prefix) throws SQLException {
		SqlFactory sf = newInstance(ds, databasename, prefix);
		sf.setAsSupportAllDatabases(true);
		for (String t : sf.getDatabase().getJdbcUtils().getTables()) {
			sf.fetchTable(t);
		}
		return sf;
	}

	private static SqlInfo getSqlInfo(DataSource ds, String databasename,
			String prefix) {
		return new SqlInfo(prefix, null, true, new JdbcUtils(ds, databasename));
	}

	/**
	 * 生成空的SQL工厂类实例。
	 * 
	 * @param ds
	 *            数据源
	 * @param databasename
	 *            数据库名称
	 */
	public static SqlFactory newInstance(DataSource ds, String databasename) {
		return newInstance(ds, databasename, (String) null);
	}

	/**
	 * 从XML配置文件中获取SQL工厂类实例。
	 * 
	 * @param ds
	 *            数据源
	 * @param databasename
	 *            数据库名称
	 * @param xml
	 *            XML配置文件流。
	 * @throws XmlParserException
	 */
	public static SqlFactory newInstance(DataSource ds, String databasename,
			InputStream xml) throws XmlParserException {
		return newInstance(ds, databasename, null, xml);
	}

	/**
	 * 生成空的SQL工厂类实例。
	 * 
	 * @param ds
	 *            数据源
	 * @param databasename
	 *            数据库名称
	 * @param prefix
	 *            数据库内SQL对象使用的前缀，该前缀用于表名，存储过程名和函数名称。
	 */
	public static SqlFactory newInstance(DataSource ds, String databasename,
			String prefix) {
		return new SqlFactory(getSqlInfo(ds, databasename, prefix));
	}

	/**
	 * 从XML配置文件中获取SQL工厂类实例。
	 * 
	 * @param ds
	 *            数据源
	 * @param databasename
	 *            数据库名称
	 * @param prefix
	 *            数据库内SQL对象使用的前缀，该前缀用于表名，存储过程名和函数名称。
	 * @param xml
	 *            XML配置文件流。
	 * @throws XmlParserException
	 */
	public static SqlFactory newInstance(DataSource ds, String databasename,
			String prefix, InputStream xml) throws XmlParserException {
		return newInstance(getSqlInfo(ds, databasename, prefix), xml);
	}

	public static SqlFactory newInstance(SqlInfo si, InputStream xml)
			throws XmlParserException {
		return SqlParser.INSTANCE.parser(xml, si);
	}

	public static <T> Mapping<T> newMapping(Class<T> cls)
			throws MappingException {
		return new MappingImpl<T>(cls);
	}

	public static <T> Mapping<T> newMapping(Class<T> cls, String text)
			throws MappingException {
		return new MappingImpl<T>(cls, text);
	}

	/**
	 * 注册某SQL工厂实例。
	 * 
	 * @param sf
	 * @param name
	 * @throws SQLExistsException
	 */
	public static void registerFactory(SqlFactory sf, String name)
			throws SQLExistsException {
		if (nmap.containsKey(name))
			throw new SQLExistsException("名称为" + name + "的SQL仓库已经存在！");
		nmap.put(name, sf);
	}

	private final SqlInfo si;
	private final SqlDatabase lib;
	private final AtomicInteger ai = new AtomicInteger();
	private final int id;
	private boolean readable = true;

	public SqlFactory(SqlInfo si) {
		this.si = si;
		this.lib = new SqlDatabase(si, si.getPrefix(), si.getName(),
				si.isMask(), null, null, null);
		id = sfid.getAndIncrement();
		map.put(id, this);
	}

	/**
	 * 创建自定义函数。
	 * 
	 * @param name
	 *            函数名
	 * @param out
	 *            返回值列表
	 * @param sql
	 *            函数体。
	 * @throws SQLExistsException
	 */
	public SqlFunction createFunction(String name, SqlVar out, String sql)
			throws SQLExistsException {
		return createFunction(null, name, true, null, null, out, sql);
	}

	/**
	 * 创建自定义函数。
	 * 
	 * @param prefix
	 *            默认前缀
	 * @param name
	 *            函数名
	 * @param ismask
	 *            是否屏蔽
	 * @param sv
	 *            参数列表
	 * @param st
	 *            调用表格列表
	 * @param out
	 *            返回值列表
	 * @param sql
	 *            函数体。
	 * @throws SQLExistsException
	 */
	public SqlFunction createFunction(String prefix, String name,
			boolean ismask, SqlVar[] sv, SqlTable[] st, SqlVar out, String sql)
			throws SQLExistsException {
		try {
			this.getDatabase().getFunction(name);
			throw new SQLExistsException("函数" + name + "已经存在！");
		} catch (SQLNotExistsException e) {
			SqlFunction f = new SqlFunction(si, prefix, name, ismask, sv, st,
					out, sql);
			this.getDatabase().addFunction(f);
			return f;
		}
	}

	/**
	 * 创建自定义存储过程。
	 * 
	 * @param name
	 *            函数名
	 * @param in
	 *            参数列表
	 * @param out
	 *            返回值列表
	 * @param sql
	 *            函数体。
	 * @throws SQLExistsException
	 */
	public SqlProcedure createProcedure(String name, SqlVar[] in, String sql,
			SqlVar... out) throws SQLExistsException {
		return createProcedure(null, name, true, in, null, null, sql, out);
	}

	/**
	 * 创建自定义存储过程。
	 * 
	 * @param prefix
	 *            默认前缀
	 * @param name
	 *            函数名
	 * @param ismask
	 *            是否屏蔽
	 * @param sv
	 *            参数列表
	 * @param st
	 *            调用表格列表
	 * @param sf
	 *            调用函数列表
	 * @param out
	 *            返回值列表
	 * @param sql
	 *            函数体。
	 * @throws SQLExistsException
	 */
	public SqlProcedure createProcedure(String prefix, String name,
			boolean ismask, SqlVar[] sv, SqlTable[] st, SqlFunction[] sf,
			String sql, SqlVar... out) throws SQLExistsException {
		try {
			this.getDatabase().getProcedure(name);
			throw new SQLExistsException("存储过程" + name + "已经存在！");
		} catch (SQLNotExistsException e) {
			SqlProcedure p = new SqlProcedure(si, prefix, name, ismask, sv, st,
					sf, sql, out);
			this.getDatabase().addProcedure(p);
			return p;
		}
	}

	/**
	 * 创建表格。
	 * 
	 * @param name
	 *            表名
	 * @param sv
	 *            字段列表。
	 * @throws SQLExistsException
	 */
	public SqlTable createTable(String name, SqlVar... sv)
			throws SQLExistsException {
		return createTable(null, name, true, sv);
	}

	/**
	 * 创建表格。
	 * 
	 * @param prefix
	 *            默认前缀
	 * @param name
	 *            表名
	 * @param ismask
	 *            是否屏蔽
	 * @param sv
	 *            字段列表。
	 * @throws SQLExistsException
	 */
	public SqlTable createTable(String prefix, String name, boolean ismask,
			SqlVar... sv) throws SQLExistsException {
		try {
			this.getDatabase().getTable(name);
			throw new SQLExistsException("表格" + name + "已经存在！");
		} catch (SQLNotExistsException e) {
			SqlTable t = new SqlTable(si, prefix, name, ismask, sv);
			this.getDatabase().addTable(t);
			return t;
		}
	}

	/**
	 * 创建格式。
	 * 
	 * @param type
	 *            Jdbc格式名称
	 */
	public SqlType createType(SqlVarType type) {
		return createType(type, 0, null);
	}

	/**
	 * 创建格式。
	 * 
	 * @param type
	 *            Jdbc格式名称
	 * @param length
	 *            长度
	 */
	public SqlType createType(SqlVarType type, int length) {
		return createType(type, length, null);
	}

	/**
	 * 创建格式。
	 * 
	 * @param type
	 *            Jdbc格式名称
	 * @param length
	 *            长度
	 * @param defaultvalue
	 *            默认值
	 */
	public SqlType createType(SqlVarType type, int length, String defaultvalue) {
		try {
			return createType(type, length, defaultvalue, -1);
		} catch (SQLException e) {
			return null;
		}
	}

	/**
	 * 创建格式。
	 * 
	 * @param name
	 *            Jdbc格式名称
	 * @param len
	 *            长度
	 * @param def
	 *            默认值
	 * @param id
	 *            序号。
	 * @throws SQLExistsException
	 */
	public SqlType createType(SqlVarType name, int len, String def, int id)
			throws SQLExistsException {
		SqlType t;
		if (id < 0) {
			id = getNewTypeIndex();
		} else {
			try {
				throw new SQLExistsException("序列号为" + id + "的字段"
						+ this.getDatabase().getType(id).toString() + "已经存在。");
			} catch (SQLNotExistsException e) {
			}
		}
		t = new SqlType(si, name, len, def, id);
		for (SqlType st : this.getDatabase().getTypes()) {
			if (st.equals(t)) {
				/**
				 * 把未使用的Type ID还原。
				 */
				ai.set(id);
				return st;
			}
		}
		this.getDatabase().addType(t);
		return t;
	}

	/**
	 * 创建字段。
	 * 
	 * @param type
	 *            格式
	 * @param name
	 *            字段名
	 */
	public SqlVar createVar(String name, SqlType type) {
		return createVar(name, type, null);
	}

	/**
	 * 创建字段。
	 * 
	 * @param type
	 *            格式
	 * @param name
	 *            字段名
	 * @param defaultvalue
	 *            默认值
	 */
	public SqlVar createVar(String name, SqlType type, String defaultvalue) {
		return createVar(name, type, defaultvalue, true);
	}

	/**
	 * 创建字段。
	 * 
	 * @param type
	 *            格式
	 * @param name
	 *            字段名
	 * @param defaultvalue
	 *            默认值
	 * @param ismask
	 *            是否屏蔽
	 */
	public SqlVar createVar(String name, SqlType type, String defaultvalue,
			boolean ismask) {
		SqlVar v = new SqlVar(type, defaultvalue, si, name, ismask);
		try {
			this.getDatabase().addVar(v);
			return v;
		} catch (SQLExistsException e) {
			try {
				return this.getDatabase().getVar(v.getId());
			} catch (SQLNotExistsException e1) {
				e1.printStackTrace();
				return null;
			}
		}
	}

	/**
	 * 从数据库中抓取某张表格,生成{@link SqlTable}，如果在工厂类实例中已经存在该表格对应的SqlTable对象，则直接返回。
	 * 
	 * @param name
	 * @throws SQLException
	 */
	public SqlTable fetchTable(String name) throws SQLException {
		// System.out.println(name);
		try {
			return this.getDatabase().getTable(name);
		} catch (SQLNotExistsException e) {
			return si.getJdbcUtils().queryMetaData(new FetchTable(this, name));
		}
	}

	/**
	 * 从数据库中抓取某张表格，并生成新的一个{@link SqlTable}对象。
	 * 
	 * @param name
	 * @param newname
	 * @throws SQLException
	 */
	public SqlTable fetchTable(String name, String newname) throws SQLException {
		try {
			this.getDatabase().getTable(newname);
			throw new SQLException("表格" + newname + "已经存在！");
		} catch (SQLNotExistsException e) {
			fetchTable(name);
			return si.getJdbcUtils().queryMetaData(
					new FetchTable(this, name, newname));
		}
	}

	/**
	 * 获取该工厂方法中的唯一SQL仓库实例。
	 */
	public SqlDatabase getDatabase() {
		return lib;
	}

	/**
	 * 该SQL工厂的识别码。
	 */
	public int getId() {
		return id;
	}

	/**
	 * 获取JDBC操作对象。
	 */
	public JdbcUtils getJdbcUtils() {
		return si.getJdbcUtils();
	}

	private int getNewTypeIndex() {
		int i = ai.getAndIncrement();
		try {
			this.getDatabase().getType(i);
			return getNewTypeIndex();
		} catch (SQLNotExistsException e) {
			return i;
		}
	}

	/**
	 * 获取SQL仓库基本信息。
	 */
	public SqlInfo getSqlInfo() {
		return si;
	}

	/**
	 * 此处显示导出的xml文档是否可读。
	 */
	public boolean isReadable() {
		return readable;
	}

	/**
	 * 导出该SQL工厂到XML流。
	 * 
	 * @param xml
	 * @throws XmlParserException
	 */
	public void save(OutputStream xml) throws XmlParserException {
		SqlParser.INSTANCE.store(this, xml);
	}

	/**
	 * 导出该SQL工厂中的所有{@link SqlTable}对象对应的数据表的创建语句。
	 * 
	 * @param dialect
	 *            选择数据库模型。
	 * @param os
	 *            输出流。
	 * @throws SqlFormatException
	 * @throws IOException
	 */
	public void saveTableCreateSql(Dialect dialect, OutputStream os)
			throws SqlFormatException, IOException {
		OutputStreamWriter osw = new OutputStreamWriter(os);
		for (SqlTable t : this.getDatabase().getTables()) {
			osw.write(dialect.getSqlDialect().getCreateTableSql(t) + "\n\n");
		}
		osw.close();
	}

	/**
	 * 设置该工厂的产品是否为全SQL平台支持。
	 * 
	 * @param issupport
	 */
	public void setAsSupportAllDatabases(boolean issupport) {
		this.getDatabase().getSqlInfo().setAsSupportAllDatabases(issupport);
	}

	/**
	 * 设置文档是否可读。
	 * 
	 * @param readable
	 */
	public void setReadable(boolean readable) {
		this.readable = readable;
	}
}
